@@ -28,7 +28,7 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #endif
 
 #include "wget.h"
-#include "md5.h"
+#include "gen-md5.h"
 
 /* Dictionary for integer-word translations.  */
 static char Wp[2048][4] = {
@@ -2152,16 +2152,16 @@ calculate_skey_response (int sequence, const char *seed, const char *pass)
   char key[8];
   static char buf[33];
 
-  struct md5_ctx ctx;
+  ALLOCA_MD5_CONTEXT (ctx);
   unsigned long results[4];	/* #### this looks 32-bit-minded */
   char *feed = (char *) alloca (strlen (seed) + strlen (pass) + 1);
 
   strcpy (feed, seed);
   strcat (feed, pass);
 
-  md5_init_ctx (&ctx);
-  md5_process_bytes (feed, strlen (feed), &ctx);
-  md5_finish_ctx (&ctx, results);
+  gen_md5_init (ctx);
+  gen_md5_update ((unsigned char *)feed, strlen (feed), ctx);
+  gen_md5_finish (ctx, (unsigned char *)results);
 
   results[0] ^= results[2];
   results[1] ^= results[3];
@@ -2169,9 +2169,9 @@ calculate_skey_response (int sequence, const char *seed, const char *pass)
 
   while (0 < sequence--)
     {
-      md5_init_ctx (&ctx);
-      md5_process_bytes (key, 8, &ctx);
-      md5_finish_ctx (&ctx, results);
+      gen_md5_init (ctx);
+      gen_md5_update ((unsigned char *)key, 8, ctx);
+      gen_md5_finish (ctx, (unsigned char *)results);
       results[0] ^= results[2];
       results[1] ^= results[3];
       memcpy (key, (char *) results, 8);
