@@ -33,9 +33,6 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include <sys/types.h>
 #include <assert.h>
 #include <errno.h>
-#ifndef WINDOWS
-# include <netdb.h>		/* for h_errno */
-#endif
 
 #include "wget.h"
 #include "utils.h"
@@ -51,17 +48,24 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #ifndef errno
 extern int errno;
 #endif
-#ifndef h_errno
-# ifndef __CYGWIN__
-extern int h_errno;
-# endif
-#endif
 
 /* File where the "ls -al" listing will be saved.  */
 #define LIST_FILENAME ".listing"
 
 extern char ftp_last_respline[];
 
+typedef struct
+{
+  int st;			/* connection status */
+  int cmd;			/* command code */
+  struct rbuf rbuf;		/* control connection buffer */
+  long dltime;			/* time of the download */
+  enum stype rs;		/* remote system reported by ftp server */ 
+  char *id;			/* initial directory */
+  char *target;			/* target file name */
+} ccon;
+
+
 /* Look for regexp "( *[0-9]+ *byte" (literal parenthesis) anywhere in
    the string S, and return the number converted to long, if found, 0
    otherwise.  */
@@ -108,21 +112,21 @@ ftp_expected_bytes (const char *s)
    connection to the server.  It always closes the data connection,
    and closes the control connection in case of error.  */
 static uerr_t
-getftp (struct urlinfo *u, long *len, long restval, ccon *con)
+getftp (struct url *u, long *len, long restval, ccon *con)
 {
   int csock, dtsock, res;
   uerr_t err;
   FILE *fp;
   char *user, *passwd, *respline;
   char *tms, *tmrate;
-  struct wget_timer *timer;
   unsigned char pasv_addr[6];
   int cmd = con->cmd;
   int passive_mode_open = 0;
   long expected_bytes = 0L;
 
   assert (con != NULL);
-  assert (u->local != NULL);
+  assert (con->target != NULL);
+
   /* Debug-check of the sanity of the request by making sure that LIST
      and RETR are never both requested (since we can handle only one
      at a time.  */
@@ -144,53 +148,34 @@ getftp (struct urlinfo *u, long *len, long restval, ccon *con)
     csock = RBUF_FD (&con->rbuf);
   else				/* cmd & DO_LOGIN */
     {
+      char type_char;
+      struct address_list *al;
+
       /* Login to the server: */
 
       /* First: Establish the control connection.  */
-      logprintf (LOG_VERBOSE, _("Connecting to %s:%hu... "), u->host, u->port);
-      err = make_connection (&csock, u->host, u->port);
+
+      al = lookup_host (u->host, 0);
+      if (!al)
+	return HOSTERR;
+      set_connection_host_name (u->host);
+      csock = connect_to_many (al, u->port, 0);
+      set_connection_host_name (NULL);
+      address_list_release (al);
+
+      if (csock < 0)
+	return errno == ECONNREFUSED ? CONREFUSED : CONERROR;
+
       if (cmd & LEAVE_PENDING)
 	rbuf_initialize (&con->rbuf, csock);
       else
 	rbuf_uninitialize (&con->rbuf);
-      switch (err)
-	{
-	  /* Do not close the socket in first several cases, since it
-	     wasn't created at all.  */
-	case HOSTERR:
-	  logputs (LOG_VERBOSE, "\n");
-	  logprintf (LOG_NOTQUIET, "%s: %s\n", u->host, herrmsg (h_errno));
-	  return HOSTERR;
-	  break;
-	case CONSOCKERR:
-	  logputs (LOG_VERBOSE, "\n");
-	  logprintf (LOG_NOTQUIET, "socket: %s\n", strerror (errno));
-	  return CONSOCKERR;
-	  break;
-	case CONREFUSED:
-	  logputs (LOG_VERBOSE, "\n");
-	  logprintf (LOG_NOTQUIET, _("Connection to %s:%hu refused.\n"),
-		     u->host, u->port);
-	  CLOSE (csock);
-	  rbuf_uninitialize (&con->rbuf);
-	  return CONREFUSED;
-	case CONERROR:
-	  logputs (LOG_VERBOSE, "\n");
-	  logprintf (LOG_NOTQUIET, "connect: %s\n", strerror (errno));
-	  CLOSE (csock);
-	  rbuf_uninitialize (&con->rbuf);
-	  return CONERROR;
-	  break;
-	default:
-	  DO_NOTHING;
-	  /* #### Hmm?  */
-	}
+
       /* Since this is a new connection, we may safely discard
 	 anything left in the buffer.  */
       rbuf_discard (&con->rbuf);
 
       /* Second: Login with proper USER/PASS sequence.  */
-      logputs (LOG_VERBOSE, _("connected!\n"));
       logprintf (LOG_VERBOSE, _("Logging in as %s ... "), user);
       if (opt.server_response)
 	logputs (LOG_ALWAYS, "\n");
@@ -271,7 +256,7 @@ Error in server response, closing control connection.\n"));
 	  abort ();
 	  break;
 	}
-      if (!opt.server_response)
+      if (!opt.server_response && err != FTPSRVERR)
 	logputs (LOG_VERBOSE, _("done.    "));
 
       /* Fourth: Find the initial ftp directory */
@@ -283,7 +268,6 @@ Error in server response, closing control connection.\n"));
       switch (err)
 	{
 	case FTPRERR:
-	case FTPSRVERR :
 	  logputs (LOG_VERBOSE, "\n");
 	  logputs (LOG_NOTQUIET, _("\
 Error in server response, closing control connection.\n"));
@@ -291,6 +275,11 @@ Error in server response, closing control connection.\n"));
 	  rbuf_uninitialize (&con->rbuf);
 	  return err;
 	  break;
+	case FTPSRVERR :
+	  /* PWD unsupported -- assume "/". */
+	  FREE_MAYBE (con->id);
+	  con->id = xstrdup ("/");
+	  break;
 	case FTPOK:
 	  /* Everything is OK.  */
 	  break;
@@ -325,9 +314,10 @@ Error in server response, closing control connection.\n"));
 	logputs (LOG_VERBOSE, _("done.\n"));
 
       /* Fifth: Set the FTP type.  */
+      type_char = ftp_process_type (u->params);
       if (!opt.server_response)
-	logprintf (LOG_VERBOSE, "==> TYPE %c ... ", TOUPPER (u->ftp_type));
-      err = ftp_type (&con->rbuf, TOUPPER (u->ftp_type));
+	logprintf (LOG_VERBOSE, "==> TYPE %c ... ", type_char);
+      err = ftp_type (&con->rbuf, type_char);
       /* FTPRERR, WRITEFAILED, FTPUNKNOWNTYPE */
       switch (err)
 	{
@@ -351,7 +341,7 @@ Error in server response, closing control connection.\n"));
 	  logputs (LOG_VERBOSE, "\n");
 	  logprintf (LOG_NOTQUIET,
 		     _("Unknown type `%c', closing control connection.\n"),
-		     TOUPPER (u->ftp_type));
+		     type_char);
 	  CLOSE (csock);
 	  rbuf_uninitialize (&con->rbuf);
 	  return err;
@@ -462,14 +452,43 @@ Error in server response, closing control connection.\n"));
   else /* do not CWD */
     logputs (LOG_VERBOSE, _("==> CWD not required.\n"));
 
+  if ((cmd & DO_RETR) && restval && *len == 0)
+    {
+      if (opt.verbose)
+	{
+          if (!opt.server_response)
+	    logprintf (LOG_VERBOSE, "==> SIZE %s ... ", u->file);
+	}
+
+      err = ftp_size(&con->rbuf, u->file, len);
+      /* FTPRERR */
+      switch (err)
+	{
+	case FTPRERR:
+	case FTPSRVERR :
+	  logputs (LOG_VERBOSE, "\n");
+	  logputs (LOG_NOTQUIET, _("\
+Error in server response, closing control connection.\n"));
+	  CLOSE (csock);
+	  rbuf_uninitialize (&con->rbuf);
+	  return err;
+	  break;
+	case FTPOK:
+	  /* Everything is OK.  */
+	  break;
+	default:
+	  abort ();
+	  break;
+	}
+	if (!opt.server_response)
+	  logputs (LOG_VERBOSE, _("done.\n"));
+    }
+
   /* If anything is to be retrieved, PORT (or PASV) must be sent.  */
   if (cmd & (DO_LIST | DO_RETR))
     {
       if (opt.ftp_pasv > 0)
 	{
-	  char thost[256];
-	  unsigned short tport;
-
 	  if (!opt.server_response)
 	    logputs (LOG_VERBOSE, "==> PASV ... ");
 	  err = ftp_pasv (&con->rbuf, pasv_addr);
@@ -509,53 +528,23 @@ Error in server response, closing control connection.\n"));
 	    }
 	  if (err==FTPOK)
 	    {
-	      sprintf (thost, "%d.%d.%d.%d",
-		       pasv_addr[0], pasv_addr[1], pasv_addr[2], pasv_addr[3]);
+	      unsigned short tport;
+
 	      tport = (pasv_addr[4] << 8) + pasv_addr[5];
-	      DEBUGP ((_("Will try connecting to %s:%hu.\n"), thost, tport));
-	      err = make_connection (&dtsock, thost, tport);
-	      switch (err)
+	      dtsock = connect_to_one (pasv_addr, tport, 1);
+
+	      if (dtsock < 0)
 		{
-		  /* Do not close the socket in first several cases,
-		     since it wasn't created at all.  */
-		case HOSTERR:
-		  logputs (LOG_VERBOSE, "\n");
-		  logprintf (LOG_NOTQUIET, "%s: %s\n", thost,
-			     herrmsg (h_errno));
+		  int save_errno = errno;
 		  CLOSE (csock);
 		  rbuf_uninitialize (&con->rbuf);
-		  return HOSTERR;
-		  break;
-		case CONSOCKERR:
-		  logputs (LOG_VERBOSE, "\n");
-		  logprintf (LOG_NOTQUIET, "socket: %s\n", strerror (errno));
-		  CLOSE (csock);
-		  rbuf_uninitialize (&con->rbuf);
-		  return CONSOCKERR;
-		  break;
-		case CONREFUSED:
-		  logputs (LOG_VERBOSE, "\n");
-		  logprintf (LOG_NOTQUIET,
-			     _("Connection to %s:%hu refused.\n"),
-			     thost, tport);
-		  CLOSE (csock);
-		  rbuf_uninitialize (&con->rbuf);
-		  closeport (dtsock);
-		  return CONREFUSED;
-		case CONERROR:
-		  logputs (LOG_VERBOSE, "\n");
-		  logprintf (LOG_NOTQUIET, "connect: %s\n",
-			     strerror (errno));
-		  CLOSE (csock);
-		  rbuf_uninitialize (&con->rbuf);
-		  closeport (dtsock);
-		  return CONERROR;
-		  break;
-		default:
-		  /* #### What?!  */
-		  DO_NOTHING;
+		  logprintf (LOG_VERBOSE, _("couldn't connect to %s:%hu: %s\n"),
+			     pretty_print_address (pasv_addr), tport,
+			     strerror (save_errno));
+		  return save_errno == ECONNREFUSED ? CONREFUSED : CONERROR;
 		}
-	      passive_mode_open= 1;  /* Flag to avoid accept port */
+
+	      passive_mode_open = 1;  /* Flag to avoid accept port */
 	      if (!opt.server_response)
 		logputs (LOG_VERBOSE, _("done.    "));
 	    } /* err==FTP_OK */
@@ -604,15 +593,6 @@ Error in server response, closing control connection.\n"));
 	      closeport (dtsock);
 	      return err;
 	      break;
-	    case HOSTERR:
-	      logputs (LOG_VERBOSE, "\n");
-	      logprintf (LOG_NOTQUIET, "%s: %s\n", u->host,
-			 herrmsg (h_errno));
-	      CLOSE (csock);
-	      closeport (dtsock);
-	      rbuf_uninitialize (&con->rbuf);
-	      return HOSTERR;
-	      break;
 	    case FTPPORTERR:
 	      logputs (LOG_VERBOSE, "\n");
 	      logputs (LOG_NOTQUIET, _("Invalid PORT.\n"));
@@ -669,7 +649,7 @@ Error in server response, closing control connection.\n"));
 	    {
 	      logprintf (LOG_NOTQUIET,
 			 _("\nREST failed; will not truncate `%s'.\n"),
-			 u->local);
+			 con->target);
 	      CLOSE (csock);
 	      closeport (dtsock);
 	      rbuf_uninitialize (&con->rbuf);
@@ -818,16 +798,16 @@ Error in server response, closing control connection.\n"));
   /* Open the file -- if opt.dfp is set, use it instead.  */
   if (!opt.dfp || con->cmd & DO_LIST)
     {
-      mkalldirs (u->local);
+      mkalldirs (con->target);
       if (opt.backups)
-	rotate_backups (u->local);
+	rotate_backups (con->target);
       /* #### Is this correct? */
-      chmod (u->local, 0600);
+      chmod (con->target, 0600);
 
-      fp = fopen (u->local, restval ? "ab" : "wb");
+      fp = fopen (con->target, restval ? "ab" : "wb");
       if (!fp)
 	{
-	  logprintf (LOG_NOTQUIET, "%s: %s\n", u->local, strerror (errno));
+	  logprintf (LOG_NOTQUIET, "%s: %s\n", con->target, strerror (errno));
 	  CLOSE (csock);
 	  rbuf_uninitialize (&con->rbuf);
 	  closeport (dtsock);
@@ -870,13 +850,12 @@ Error in server response, closing control connection.\n"));
 		   legible (expected_bytes - restval));
       logputs (LOG_VERBOSE, _(" (unauthoritative)\n"));
     }
-  timer = wtimer_new ();
+
   /* Get the contents of the document.  */
-  res = get_contents (dtsock, fp, len, restval, expected_bytes, &con->rbuf, 0);
-  con->dltime = wtimer_elapsed (timer);
-  wtimer_delete (timer);
+  res = get_contents (dtsock, fp, len, restval, expected_bytes, &con->rbuf,
+		      0, &con->dltime);
   tms = time_str (NULL);
-  tmrate = rate (*len - restval, con->dltime, 0);
+  tmrate = retr_rate (*len - restval, con->dltime, 0);
   /* Close data connection socket.  */
   closeport (dtsock);
   /* Close the local file.  */
@@ -896,7 +875,7 @@ Error in server response, closing control connection.\n"));
   if (res == -2)
     {
       logprintf (LOG_NOTQUIET, _("%s: %s, closing control connection.\n"),
-		 u->local, strerror (errno));
+		 con->target, strerror (errno));
       CLOSE (csock);
       rbuf_uninitialize (&con->rbuf);
       return FWRITEERR;
@@ -961,10 +940,10 @@ Error in server response, closing control connection.\n"));
      print it out.  */
   if (opt.server_response && (con->cmd & DO_LIST))
     {
-      mkalldirs (u->local);
-      fp = fopen (u->local, "r");
+      mkalldirs (con->target);
+      fp = fopen (con->target, "r");
       if (!fp)
-	logprintf (LOG_ALWAYS, "%s: %s\n", u->local, strerror (errno));
+	logprintf (LOG_ALWAYS, "%s: %s\n", con->target, strerror (errno));
       else
 	{
 	  char *line;
@@ -988,7 +967,7 @@ Error in server response, closing control connection.\n"));
    This loop either gets commands from con, or (if ON_YOUR_OWN is
    set), makes them up to retrieve the file given by the URL.  */
 static uerr_t
-ftp_loop_internal (struct urlinfo *u, struct fileinfo *f, ccon *con)
+ftp_loop_internal (struct url *u, struct fileinfo *f, ccon *con)
 {
   int count, orig_lp;
   long restval, len;
@@ -996,21 +975,21 @@ ftp_loop_internal (struct urlinfo *u, struct fileinfo *f, ccon *con)
   uerr_t err;
   struct stat st;
 
-  if (!u->local)
-    u->local = url_filename (u);
+  if (!con->target)
+    con->target = url_filename (u);
 
-  if (opt.noclobber && file_exists_p (u->local))
+  if (opt.noclobber && file_exists_p (con->target))
     {
       logprintf (LOG_VERBOSE,
-		 _("File `%s' already there, not retrieving.\n"), u->local);
+		 _("File `%s' already there, not retrieving.\n"), con->target);
       /* If the file is there, we suppose it's retrieved OK.  */
       return RETROK;
     }
 
   /* Remove it if it's a link.  */
-  remove_link (u->local);
+  remove_link (con->target);
   if (!opt.output_document)
-    locf = u->local;
+    locf = con->target;
   else
     locf = opt.output_document;
 
@@ -1068,7 +1047,7 @@ ftp_loop_internal (struct urlinfo *u, struct fileinfo *f, ccon *con)
       /* Print fetch message, if opt.verbose.  */
       if (opt.verbose)
 	{
-	  char *hurl = str_url (u->proxy ? u->proxy : u, 1);
+	  char *hurl = url_string (u, 1);
 	  char tmp[15];
 	  strcpy (tmp, "        ");
 	  if (count > 1)
@@ -1125,7 +1104,7 @@ ftp_loop_internal (struct urlinfo *u, struct fileinfo *f, ccon *con)
 	}
       /* Time?  */
       tms = time_str (NULL);
-      tmrate = rate (len - restval, con->dltime, 0);
+      tmrate = retr_rate (len - restval, con->dltime, 0);
 
       /* If we get out of the switch above without continue'ing, we've
 	 successfully downloaded a file.  Remember this fact. */
@@ -1143,7 +1122,7 @@ ftp_loop_internal (struct urlinfo *u, struct fileinfo *f, ccon *con)
 	  /* Need to hide the password from the URL.  The `if' is here
              so that we don't do the needless allocation every
              time. */
-	  char *hurl = str_url (u->proxy ? u->proxy : u, 1);
+	  char *hurl = url_string (u, 1);
 	  logprintf (LOG_NONVERBOSE, "%s URL: %s [%ld] -> \"%s\" [%d]\n",
 		     tms, hurl, len, locf, count);
 	  xfree (hurl);
@@ -1203,43 +1182,48 @@ ftp_loop_internal (struct urlinfo *u, struct fileinfo *f, ccon *con)
 /* Return the directory listing in a reusable format.  The directory
    is specifed in u->dir.  */
 uerr_t
-ftp_get_listing (struct urlinfo *u, ccon *con, struct fileinfo **f)
+ftp_get_listing (struct url *u, ccon *con, struct fileinfo **f)
 {
   uerr_t err;
-  char *olocal = u->local;
-  char *list_filename, *ofile;
+  char *uf;			/* url file name */
+  char *lf;			/* list file name */
+  char *old_target = con->target;
 
   con->st &= ~ON_YOUR_OWN;
   con->cmd |= (DO_LIST | LEAVE_PENDING);
   con->cmd &= ~DO_RETR;
-  /* Get the listing filename.  */
-  ofile = u->file;
-  u->file = LIST_FILENAME;
-  list_filename = url_filename (u);
-  u->file = ofile;
-  u->local = list_filename;
-  DEBUGP ((_("Using `%s' as listing tmp file.\n"), list_filename));
+
+  /* Find the listing file name.  We do it by taking the file name of
+     the URL and replacing the last component with the listing file
+     name.  */
+  uf = url_filename (u);
+  lf = file_merge (uf, LIST_FILENAME);
+  xfree (uf);
+  DEBUGP ((_("Using `%s' as listing tmp file.\n"), lf));
+
+  con->target = lf;
   err = ftp_loop_internal (u, NULL, con);
-  u->local = olocal;
+  con->target = old_target;
+
   if (err == RETROK)
-    *f = ftp_parse_ls (list_filename, con->rs);
+    *f = ftp_parse_ls (lf, con->rs);
   else
     *f = NULL;
   if (opt.remove_listing)
     {
-      if (unlink (list_filename))
+      if (unlink (lf))
 	logprintf (LOG_NOTQUIET, "unlink: %s\n", strerror (errno));
       else
-	logprintf (LOG_VERBOSE, _("Removed `%s'.\n"), list_filename);
+	logprintf (LOG_VERBOSE, _("Removed `%s'.\n"), lf);
     }
-  xfree (list_filename);
+  xfree (lf);
   con->cmd &= ~DO_LIST;
   return err;
 }
 
-static uerr_t ftp_retrieve_dirs PARAMS ((struct urlinfo *, struct fileinfo *,
+static uerr_t ftp_retrieve_dirs PARAMS ((struct url *, struct fileinfo *,
 					 ccon *));
-static uerr_t ftp_retrieve_glob PARAMS ((struct urlinfo *, ccon *, int));
+static uerr_t ftp_retrieve_glob PARAMS ((struct url *, ccon *, int));
 static struct fileinfo *delelement PARAMS ((struct fileinfo *,
 					    struct fileinfo **));
 static void freefileinfo PARAMS ((struct fileinfo *f));
@@ -1252,11 +1236,10 @@ static void freefileinfo PARAMS ((struct fileinfo *f));
    If opt.recursive is set, after all files have been retrieved,
    ftp_retrieve_dirs will be called to retrieve the directories.  */
 static uerr_t
-ftp_retrieve_list (struct urlinfo *u, struct fileinfo *f, ccon *con)
+ftp_retrieve_list (struct url *u, struct fileinfo *f, ccon *con)
 {
   static int depth = 0;
   uerr_t err;
-  char *olocal, *ofile;
   struct fileinfo *orig;
   long local_size;
   time_t tml;
@@ -1291,15 +1274,19 @@ ftp_retrieve_list (struct urlinfo *u, struct fileinfo *f, ccon *con)
 
   while (f)
     {
+      char *old_target, *ofile;
+
       if (downloaded_exceeds_quota ())
 	{
 	  --depth;
 	  return QUOTEXC;
 	}
-      olocal = u->local;
-      ofile = u->file;
-      u->file = f->name;
-      u->local = url_filename (u);
+      old_target = con->target;
+
+      ofile = xstrdup (u->file);
+      url_set_file (u, f->name);
+
+      con->target = url_filename (u);
       err = RETROK;
 
       dlthis = 1;
@@ -1311,7 +1298,7 @@ ftp_retrieve_list (struct urlinfo *u, struct fileinfo *f, ccon *con)
 	     I'm not implementing it now since files on an FTP server are much
 	     more likely than files on an HTTP server to legitimately have a
 	     .orig suffix. */
-	  if (!stat (u->local, &st))
+	  if (!stat (con->target, &st))
 	    {
               int eq_size;
               int cor_val;
@@ -1328,7 +1315,7 @@ ftp_retrieve_list (struct urlinfo *u, struct fileinfo *f, ccon *con)
 		  /* Remote file is older, file sizes can be compared and
                      are both equal. */
                   logprintf (LOG_VERBOSE, _("\
-Remote file no newer than local file `%s' -- not retrieving.\n"), u->local);
+Remote file no newer than local file `%s' -- not retrieving.\n"), con->target);
 		  dlthis = 0;
 		}
 	      else if (eq_size)
@@ -1336,7 +1323,7 @@ Remote file no newer than local file `%s' -- not retrieving.\n"), u->local);
                   /* Remote file is newer or sizes cannot be matched */
                   logprintf (LOG_VERBOSE, _("\
 Remote file is newer than local file `%s' -- retrieving.\n\n"),
-                             u->local);
+                             con->target);
                 }
               else
                 {
@@ -1364,30 +1351,30 @@ The sizes do not match (local %ld) -- retrieving.\n\n"), local_size);
                   struct stat st;
 		  /* Check whether we already have the correct
                      symbolic link.  */
-                  int rc = lstat (u->local, &st);
+                  int rc = lstat (con->target, &st);
                   if (rc == 0)
 		    {
 		      size_t len = strlen (f->linkto) + 1;
 		      if (S_ISLNK (st.st_mode))
 			{
 			  char *link_target = (char *)alloca (len);
-			  size_t n = readlink (u->local, link_target, len);
+			  size_t n = readlink (con->target, link_target, len);
 			  if ((n == len - 1)
 			      && (memcmp (link_target, f->linkto, n) == 0))
 			    {
 			      logprintf (LOG_VERBOSE, _("\
 Already have correct symlink %s -> %s\n\n"),
-					 u->local, f->linkto);
+					 con->target, f->linkto);
                               dlthis = 0;
 			      break;
 			    }
 			}
 		    }
 		  logprintf (LOG_VERBOSE, _("Creating symlink %s -> %s\n"),
-			     u->local, f->linkto);
+			     con->target, f->linkto);
 		  /* Unlink before creating symlink!  */
-		  unlink (u->local);
-		  if (symlink (f->linkto, u->local) == -1)
+		  unlink (con->target);
+		  if (symlink (f->linkto, con->target) == -1)
 		    logprintf (LOG_NOTQUIET, "symlink: %s\n",
 			       strerror (errno));
 		  logputs (LOG_VERBOSE, "\n");
@@ -1395,7 +1382,7 @@ Already have correct symlink %s -> %s\n\n"),
 #else  /* not HAVE_SYMLINK */
 	      logprintf (LOG_NOTQUIET,
 			 _("Symlinks not supported, skipping symlink `%s'.\n"),
-			 u->local);
+			 con->target);
 #endif /* not HAVE_SYMLINK */
 	    }
 	  else                /* opt.retr_symlinks */
@@ -1426,7 +1413,7 @@ Already have correct symlink %s -> %s\n\n"),
       if (!(f->type == FT_SYMLINK && !opt.retr_symlinks)
 	  && f->tstamp != -1
           && dlthis
-	  && file_exists_p (u->local))
+	  && file_exists_p (con->target))
 	{
 	  /* #### This code repeats in http.c and ftp.c.  Move it to a
              function!  */
@@ -1437,27 +1424,31 @@ Already have correct symlink %s -> %s\n\n"),
 		fl = opt.output_document;
 	    }
 	  else
-	    fl = u->local;
+	    fl = con->target;
 	  if (fl)
 	    touch (fl, f->tstamp);
 	}
       else if (f->tstamp == -1)
-	logprintf (LOG_NOTQUIET, _("%s: corrupt time-stamp.\n"), u->local);
+	logprintf (LOG_NOTQUIET, _("%s: corrupt time-stamp.\n"), con->target);
 
       if (f->perms && f->type == FT_PLAINFILE && dlthis)
-	chmod (u->local, f->perms);
+	chmod (con->target, f->perms);
       else
-	DEBUGP (("Unrecognized permissions for %s.\n", u->local));
+	DEBUGP (("Unrecognized permissions for %s.\n", con->target));
+
+      xfree (con->target);
+      con->target = old_target;
+
+      url_set_file (u, ofile);
+      xfree (ofile);
 
-      xfree (u->local);
-      u->local = olocal;
-      u->file = ofile;
       /* Break on fatals.  */
       if (err == QUOTEXC || err == HOSTERR || err == FWRITEERR)
 	break;
       con->cmd &= ~ (DO_CWD | DO_LOGIN);
       f = f->next;
-    } /* while */
+    }
+
   /* We do not want to call ftp_retrieve_dirs here */
   if (opt.recursive &&
       !(opt.reclevel != INFINITE_RECURSION && depth >= opt.reclevel))
@@ -1474,51 +1465,62 @@ Already have correct symlink %s -> %s\n\n"),
    ftp_retrieve_glob on each directory entry.  The function knows
    about excluded directories.  */
 static uerr_t
-ftp_retrieve_dirs (struct urlinfo *u, struct fileinfo *f, ccon *con)
+ftp_retrieve_dirs (struct url *u, struct fileinfo *f, ccon *con)
 {
-  char *odir;
-  char *current_container = NULL;
-  int current_length = 0;
+  char *container = NULL;
+  int container_size = 0;
 
   for (; f; f = f->next)
     {
-      int len;
+      int size;
+      char *odir, *newdir;
 
       if (downloaded_exceeds_quota ())
 	break;
       if (f->type != FT_DIRECTORY)
 	continue;
-      odir = u->dir;
-      len = strlen (u->dir) + 1 + strlen (f->name) + 1;
+
       /* Allocate u->dir off stack, but reallocate only if a larger
-         string is needed.  */
-      if (len > current_length)
-	current_container = (char *)alloca (len);
-      u->dir = current_container;
+         string is needed.  It's a pity there's no "realloca" for an
+         item on the bottom of the stack.  */
+      size = strlen (u->dir) + 1 + strlen (f->name) + 1;
+      if (size > container_size)
+	container = (char *)alloca (size);
+      newdir = container;
+
+      odir = u->dir;
       if (*odir == '\0'
 	  || (*odir == '/' && *(odir + 1) == '\0'))
 	/* If ODIR is empty or just "/", simply append f->name to
 	   ODIR.  (In the former case, to preserve u->dir being
 	   relative; in the latter case, to avoid double slash.)  */
-	sprintf (u->dir, "%s%s", odir, f->name);
+	sprintf (newdir, "%s%s", odir, f->name);
       else
 	/* Else, use a separator. */
-	sprintf (u->dir, "%s/%s", odir, f->name);
+	sprintf (newdir, "%s/%s", odir, f->name);
+
       DEBUGP (("Composing new CWD relative to the initial directory.\n"));
-      DEBUGP (("  odir = '%s'\n  f->name = '%s'\n  u->dir = '%s'\n\n",
-	       odir, f->name, u->dir));
-      if (!accdir (u->dir, ALLABS))
+      DEBUGP (("  odir = '%s'\n  f->name = '%s'\n  newdir = '%s'\n\n",
+	       odir, f->name, newdir));
+      if (!accdir (newdir, ALLABS))
 	{
 	  logprintf (LOG_VERBOSE, _("\
-Not descending to `%s' as it is excluded/not-included.\n"), u->dir);
-	  u->dir = odir;
+Not descending to `%s' as it is excluded/not-included.\n"), newdir);
 	  continue;
 	}
+
       con->st &= ~DONE_CWD;
+
+      odir = xstrdup (u->dir);	/* because url_set_dir will free
+				   u->dir. */
+      url_set_dir (u, newdir);
       ftp_retrieve_glob (u, con, GETALL);
+      url_set_dir (u, odir);
+      xfree (odir);
+
       /* Set the time-stamp?  */
-      u->dir = odir;
     }
+
   if (opt.quota && opt.downloaded > opt.quota)
     return QUOTEXC;
   else
@@ -1535,7 +1537,7 @@ Not descending to `%s' as it is excluded/not-included.\n"), u->dir);
    get the listing, so that the time-stamp is heeded); if it's GLOBALL,
    use globbing; if it's GETALL, download the whole directory.  */
 static uerr_t
-ftp_retrieve_glob (struct urlinfo *u, ccon *con, int action)
+ftp_retrieve_glob (struct url *u, ccon *con, int action)
 {
   struct fileinfo *orig, *start;
   uerr_t res;
@@ -1575,7 +1577,7 @@ ftp_retrieve_glob (struct urlinfo *u, ccon *con, int action)
 	  matchres = fnmatch (u->file, f->name, 0);
 	  if (matchres == -1)
 	    {
-	      logprintf (LOG_NOTQUIET, "%s: %s\n", u->local,
+	      logprintf (LOG_NOTQUIET, "%s: %s\n", con->target,
 			 strerror (errno));
 	      break;
 	    }
@@ -1625,7 +1627,7 @@ ftp_retrieve_glob (struct urlinfo *u, ccon *con, int action)
    of URL.  Inherently, its capabilities are limited on what can be
    encoded into a URL.  */
 uerr_t
-ftp_loop (struct urlinfo *u, int *dt)
+ftp_loop (struct url *u, int *dt)
 {
   ccon con;			/* FTP connection */
   uerr_t res;
@@ -1654,7 +1656,7 @@ ftp_loop (struct urlinfo *u, int *dt)
 	    {
 	      char *filename = (opt.output_document
 				? xstrdup (opt.output_document)
-				: (u->local ? xstrdup (u->local)
+				: (con.target ? xstrdup (con.target)
 				   : url_filename (u)));
 	      res = ftp_index (filename, u, f);
 	      if (res == FTPOK && opt.verbose)
@@ -1704,6 +1706,8 @@ ftp_loop (struct urlinfo *u, int *dt)
     CLOSE (RBUF_FD (&con.rbuf));
   FREE_MAYBE (con.id);
   con.id = NULL;
+  FREE_MAYBE (con.target);
+  con.target = NULL;
   return res;
 }
 
