@@ -36,9 +36,6 @@ struct options
   int relative_only;		/* Follow only relative links. */
   int no_parent;		/* Restrict access to the parent
 				   directory.  */
-  int simple_check;		/* Should we use simple checking
-				   (strcmp) or do we create a host
-				   hash and call gethostbyname? */
   int reclevel;			/* Maximum level of recursion */
   int dirstruct;		/* Do we build the directory structure
 				  as we go along? */
@@ -49,7 +46,6 @@ struct options
 				   data. */
   char *dir_prefix;		/* The top of directory tree */
   char *lfilename;		/* Log filename */
-  int no_flush;			/* If non-zero, inhibit flushing log. */
   char *input_filename;		/* Input filename */
   int force_html;		/* Is the input file an HTML file? */
 
@@ -96,16 +92,20 @@ struct options
   char *http_proxy, *ftp_proxy, *https_proxy;
   char **no_proxy;
   char *base_href;
+  char *progress_type;		/* progress indicator type. */
   char *proxy_user; /*oli*/
   char *proxy_passwd;
 #ifdef HAVE_SELECT
   long timeout;			/* The value of read timeout in
 				   seconds. */
 #endif
+  int random_wait;		/* vary from 0 .. wait secs by random()? */
   long wait;			/* The wait period between retrievals. */
   long waitretry;		/* The wait period between retries. - HEH */
   int use_robots;		/* Do we heed robots.txt? */
 
+  long limit_rate;		/* Limit the download rate to this
+				   many bps. */
   long quota;			/* Maximum number of bytes to
 				   retrieve. */
   VERY_LONG_TYPE downloaded;	/* How much we downloaded already. */
@@ -139,6 +139,7 @@ struct options
   int htmlify;			/* Do we HTML-ify the OS-dependent
 				   listings? */
 
+  char *dot_style;
   long dot_bytes;		/* How many bytes in a printing
 				   dot. */
   int dots_in_line;		/* How many dots in one line. */
@@ -159,6 +160,7 @@ struct options
   char *sslcertkey;		/* the keyfile for this certificate
 				   (if not internal) included in the
 				   certfile. */
+  char *sslegdsock;             /* optional socket of the egd daemon */
 #endif /* HAVE_SSL */
 
   int   cookies;
