@@ -647,7 +647,7 @@ create_image (struct bar_progress *bp, long dltime)
 	/* Hours not printed: pad with three spaces (two digits and
 	   colon). */
 	APPEND_LITERAL ("   ");
-      else if (eta_hrs >= 10)
+      else if (eta_hrs < 10)
 	/* Hours printed with one digit: pad with one space. */
 	*p++ = ' ';
       else
@@ -683,8 +683,10 @@ create_image (struct bar_progress *bp, long dltime)
 static void
 display_image (char *buf)
 {
+  int old = log_set_save_context (0);
   logputs (LOG_VERBOSE, "\r");
   logputs (LOG_VERBOSE, buf);
+  log_set_save_context (old);
 }
 
 static void
