@@ -422,8 +422,7 @@ free_specs (struct robot_specs *specs)
 
 /* The inner matching engine: return non-zero if RECORD_PATH matches
    URL_PATH.  The rules for matching are described at
-   <http://info.webcrawler.com/mak/projects/robots/norobots-rfc.html>,
-   section 3.2.2.  */
+   <http://www.robotstxt.org/wc/norobots-rfc.txt>, section 3.2.2.  */
 
 static int
 matches (const char *record_path, const char *url_path)
@@ -478,7 +477,7 @@ static struct hash_table *registered_specs;
   result = alloca (HP_len + 1 + numdigit (port) + 1);	\
   memcpy (result, host, HP_len);			\
   result[HP_len] = ':';					\
-  long_to_string (result + HP_len + 1, port);		\
+  number_to_string (result + HP_len + 1, port);		\
 } while (0)
 
 /* Register RES specs that below to server on HOST:PORT.  They will
