@@ -22,13 +22,15 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #include "rbuf.h"
 
-int get_contents PARAMS ((int, FILE *, long *, long, long, struct rbuf *, int));
+int get_contents PARAMS ((int, FILE *, long *, long, long, struct rbuf *,
+			  int, long *));
 
 uerr_t retrieve_url PARAMS ((const char *, char **, char **,
 			     const char *, int *));
 uerr_t retrieve_from_file PARAMS ((const char *, int, int *));
 
-char *rate PARAMS ((long, long, int));
+char *retr_rate PARAMS ((long, long, int));
+double calc_rate PARAMS ((long, long, int *));
 void printwhat PARAMS ((int, int));
 
 void downloaded_increase PARAMS ((unsigned long));
@@ -36,4 +38,12 @@ int downloaded_exceeds_quota PARAMS ((void));
 
 void sleep_between_retrievals PARAMS ((int));
 
+/* Because there's no http.h. */
+
+struct url;
+
+uerr_t http_loop PARAMS ((struct url *, char **, char **, const char *,
+			  int *, struct url *));
+
+
 #endif /* RETR_H */
